@@ -8,7 +8,7 @@ Usage:
   protococo.py find   [<message_hex_string> ...]
                       [--cocofile=<file> --format=<option>]
                       [--dissect | --dissect-fields=<comma_separated_fields>]
-                      [--list --verbose --decode]
+                      [--list --verbose --decode --long-name]
   protococo.py create (<message_name> | --from-json=<json_file>)
                       [--cocofile=<file>]
   protococo.py json-recipe <message_names> ...
@@ -26,7 +26,7 @@ Options:
   
 """
 
-__version__ = "0.0.1"
+__version__ = "0.1.0"
 
 from pprint import *
 from collections import OrderedDict
@@ -34,6 +34,7 @@ import re
 import os
 import sys
 import json
+import copy
 from docopt import docopt
 
 # TODO: Esto pide a gritos como mínimo 3 clases: FieldRule (con byte_symbol, field_name y params), SpecialRule (con params), TitleRule (con message_name)
@@ -214,11 +215,15 @@ def override_rules(parent_rules, child_rules):
     #pprint(override_dict)
     
     for overriden_field_name, subfields in override_dict.items():
+        found_override = False
         for i, parent_rule in enumerate(tokenized_overriden_rules):
             #print(f"{parent_rule=}")
             if rule_is_field(parent_rule) and overriden_field_name == field_rule_get_field_name(parent_rule):
                 tokenized_overriden_rules = tokenized_overriden_rules[:i] + subfields + tokenized_overriden_rules[i+1:]
+                found_override = True
                 break
+        if not found_override:
+            raise RuntimeError(f"Overriding spec '{title_rule_get_name(child_rules[0])}', couldn't find overriden field '{overriden_field_name}' in parent spec '{parent_name}'")
     
     #pprint(tokenized_overriden_rules)
     
@@ -241,21 +246,67 @@ def perform_subtypeof_overrides(child_tokenized_rules, all_messages_rules_tokeni
             else:
                 #pprint(f"found parent {parent_rules[0]=} of subtype for {rule=}")
                 
-                #pprint(f"before overide: {expanded_child_rules=}")
-                #print(f"{expanded_child_rules[i]=}")
+                #pprint(f"before override:")
+                #pprint(expanded_child_rules)
+                # pprint(f"{expanded_child_rules[i]=}")
                 expanded_child_rules = override_rules(parent_rules, expanded_child_rules)
-                #pprint(f"after override: {expanded_child_rules=}")
-                #print(f"{expanded_child_rules[i]=}")
+                #pprint(f"after override")
+                #pprint(expanded_child_rules)
+                #pprint(f"{expanded_child_rules[i]=}")
 
             i = 0
         i = i+1
     return expanded_child_rules
 
+def preprocess_encode_fields(all_messages_rules_tokenized):
+    for message_rules in all_messages_rules_tokenized:
+        for rule in message_rules:
+            if rule_is_field(rule):
+                byte_symbol = rule[0].strip()
+                if byte_symbol[0] == "(" and byte_symbol[-1] == ")":
+                    #Override field rule byte symbol with the encoded value (hex string)
+                    rule[0] = field_encode(rule, byte_symbol[1:-1])
+
+def preprocess_multiple_subtypeof(all_messages_rules_tokenized):
+    j = 0
+    while j < len(all_messages_rules_tokenized):
+        message_rules = all_messages_rules_tokenized[j]
+        for i, rule in enumerate(message_rules):
+            if rule_is_subtypeof(rule):
+                subtypeof_rule_args_string = subtypeof_rule_get_parent(rule)
+                parents = subtypeof_rule_args_string.split()
+                if len(parents) > 1:
+                    assert rule_is_title(message_rules[0])
+
+                    new_message_mangled_names = []
+                    #then calculate rest of variations (for rest of parents) and append to the cocodocument:
+                    for parent in parents:
+                        new_message_spec = copy.deepcopy(message_rules)
+                        new_message_spec[i] = ["", f"subtypeof {parent}"] #replace subtypeof rule with a subtypeof rule - with each parent
+                        assert rule_is_title(new_message_spec[0])
+                        new_message_spec[0][0] = f"[{parent}.{new_message_spec[0][0][1:]}" #replace title of spec with mangled name #TODO can I delete this line?
+                        new_message_mangled_names.append(title_rule_get_name(new_message_spec[0]))
+
+                        all_messages_rules_tokenized = all_messages_rules_tokenized[:j] + [new_message_spec] + all_messages_rules_tokenized[j:]
+                        j+=1
+                    
+                    del all_messages_rules_tokenized[j]
+                    j-=1
+                        
+                    #now, make variations for all children
+                    for k, mr in enumerate(all_messages_rules_tokenized):
+                        for m, r in enumerate(mr):
+                            if rule_is_subtypeof(r):
+                                if title_rule_get_name(message_rules[0]) in subtypeof_rule_get_parent(r).split():
+                                    r = ["", f"subtypeof {' '.join(new_message_mangled_names)}"]
+                                    all_messages_rules_tokenized[k][m] = r
+        j+=1
+    
+    return all_messages_rules_tokenized
+
 def full_field_names_refer_to_same(a, b):
     return re.sub(":[^\.]*", "", a) == re.sub(":[^\.]*", "", b)
 
-#print(full_field_names_refer_to_same("general_message_format:data_to_component.MESSAGE BODY", "general_message_format.MESSAGE BODY"))
-
 def rule_is_multifieldstart(rule):
     return len(rule) == 2 and rule[1][:15] == "startmultifield"
 
@@ -336,8 +387,11 @@ def identify_message(message, all_messages_rules_tokenized):
                 result += 1
         
         return result
+    
+    def number_of_parents(message_name):
+        return len(get_subtype_parents(message_name, all_messages_rules_tokenized, False))
 
-    ordered_message_names = sorted(message_names, key=lambda x: (total_bytes_matching(validate_results[x]), total_fields_matching(validate_results[x])), reverse=True)
+    ordered_message_names = sorted(message_names, key=lambda x: (total_bytes_matching(validate_results[x]), total_fields_matching(validate_results[x]), number_of_parents(x)), reverse=True)
 
     #return True, ordered_message_names
     return ordered_message_names, validate_results
@@ -533,7 +587,7 @@ def field_get_expected_bytes_length(field_rule, previous_length_params_list, act
             
             #print(f"Checking in not-multifields, {field_name=}, {get_field_name_from_length_param(param)=}")
             
-            if field_name == get_field_name_from_length_param(param):
+            if field_name == get_field_name_from_length_param(param) and unmatched_multifieldstart_params_to_stack(active_multifields) == param["multifields_stack"]:
                 foundLength = True
                 return get_length_from_length_param(param, message)
         
@@ -580,16 +634,36 @@ def field_get_expected_bytes_length(field_rule, previous_length_params_list, act
     if not foundLength:
         raise RuntimeError(f"Length of N field not found in previous fields for rule: {field_rule}")
 
+def unmatched_multifieldstart_params_to_stack(unmatched_multifieldstart_params):
+    return [get_multifieldstart_full_name(i["param"]) for i in unmatched_multifieldstart_params]
+
+def get_conflicting_field_names(tokenized_message_rules):
+    conflicting_names = []
+    names = []
+    for rule in tokenized_message_rules:
+        if rule_is_field(rule):
+            field_name = rule[1]
+            if field_name in names:
+                if not field_name in conflicting_names:
+                    conflicting_names.append(field_name)
+            else:
+                names.append(field_name)
+
+    return conflicting_names
 
 def validate_message(message_rules, message, all_messages_rules_tokenized):
     assert is_valid_message_input(message), "Malformed message, invalid hex string"
     
     tokenized_rules = tokenize_rules(message_rules) if isinstance(message_rules, str) else message_rules
-    #pprint(f"before overriding: {tokenized_rules=}")
+    #print("before overriding:")
+    #pprint(tokenized_rules)
 
     tokenized_rules = perform_subtypeof_overrides(tokenized_rules, all_messages_rules_tokenized)
     ## FROM THIS POINT ON WE HAVE OUR TOKENIZED RULES FULLY EXPANDED
+    #print("after overriding:")
     #pprint(tokenized_rules)
+
+    conflicting_field_names = get_conflicting_field_names(tokenized_rules)
     
     title_rule = None
     title = None
@@ -640,8 +714,9 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                 pprint(e)
             
             
-            expected_lengths_dict[field_name] = current_length
+            expected_lengths_dict[(tuple(unmatched_multifieldstart_params_to_stack(unmatched_multifieldstart_params)), field_name)] = current_length
             
+
             if byte_symbol_is_valid_hex(byte_symbol):
                 pass
             elif byte_symbol_is_XX_rule_type(byte_symbol):
@@ -653,7 +728,8 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                             "param" : param, 
                             "value_offset" : current_offset, 
                             "value_length" : current_length,
-                            "parent_message_name" : title
+                            "parent_message_name" : title,
+                            "multifields_stack" : unmatched_multifieldstart_params_to_stack(unmatched_multifieldstart_params),
                         })
                         #print(length_params)
             elif byte_symbol == "N":
@@ -668,41 +744,48 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
     i = 0
     current_offset = 0
     current_length = 0
+
+    multifield_names_stack = []
     while i < len(tokenized_rules):
             
         rule = tokenized_rules[i]
-        
+
         if rule_is_field(rule):
             
             byte_symbol = rule[0]
             field_name = rule[1]
             params = rule[2:]
-            
+
+            field_name_mangled = field_name if not field_name in conflicting_field_names else ".".join([i.split(".", 1)[1] for i in multifield_names_stack] + [field_name])
+
+            if field_name_mangled in result_dict.keys():
+                raise RuntimeError(f"Duplicated field {field_name} in {rule=} of {title_rule_get_name(title_rule)}")
+
             if field_name in unknown_length_fields:
-                result_dict[field_name] = "(?)"
-                diff_dict[field_name]  = False
+                result_dict[field_name_mangled] = "(?)"
+                diff_dict[field_name_mangled]  = False
                 log_message = "Can't deduce this field length"
                 try:
-                    log_dict[field_name].append(log_message)
+                    log_dict[field_name_mangled].append(log_message)
                 except KeyError:
-                    log_dict[field_name] = [log_message]
+                    log_dict[field_name_mangled] = [log_message]
                 i+=1
                 continue
             
-            current_length = expected_lengths_dict[field_name]
+            current_length = expected_lengths_dict[(tuple(multifield_names_stack), field_name)]
             
             #print(f"{current_offset=},{current_length=}, '{message[current_offset:current_offset + current_length * 2]}'")
             
             if not "." in message[current_offset:current_offset + current_length * 2] and current_offset + current_length * 2 > len(message):
-                result_dict[field_name] = message[current_offset:]
+                result_dict[field_name_mangled] = message[current_offset:]
                 
-                number_of_missing_bytes = current_length - len(result_dict[field_name])//2
+                number_of_missing_bytes = current_length - len(result_dict[field_name_mangled])//2
                 if number_of_missing_bytes <= 16:
-                    result_dict[field_name] += "--" * number_of_missing_bytes
+                    result_dict[field_name_mangled] += "--" * number_of_missing_bytes
                 else:
-                    result_dict[field_name] += f"---({number_of_missing_bytes} bytes missing)---"
+                    result_dict[field_name_mangled] += f"---({number_of_missing_bytes} bytes missing)---"
                 
-                diff_dict[field_name] = False
+                diff_dict[field_name_mangled] = False
                 current_offset = len(message)
             
             elif "." in message[current_offset:current_offset + current_length * 2]: # and not message[current_offset:current_offset+3] == "...":
@@ -714,19 +797,29 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                 this_field_expected_length = 0
                 total_expected_length = 0
                 expected_length_after_this_field = 0
+                
+                mfns_stack = []
                 for j, r in enumerate(tokenized_rules):
                     if (rule_is_field(r)):
-                        r_length = expected_lengths_dict[field_rule_get_field_name(r)]
+                        r_length = expected_lengths_dict[(tuple(mfns_stack),field_rule_get_field_name(r))]
                         total_expected_length += r_length
                         if j==i:
                             this_field_expected_length = r_length
                         elif j>i:
                             expected_length_after_this_field += r_length
+                    elif rule_is_multifieldstart(r):
+                        multifield_name = get_multifieldstart_full_name(r[1])
+                        mfns_stack.append(multifield_name)
+                    elif rule_is_multifieldend(r):
+                        mfs_name = mfns_stack[-1]
+                        mfe_name = get_multifieldend_full_name(r[1])
+                        assert mfs_name == mfe_name, f"Unexpected multifield end, {r=}, {mfns_stack[-1]=}"
+                        mfns_stack.pop()
                 #print(f"{total_expected_length=}")
                 #print(f"{expected_length_after_this_field=}")
 
-                result_dict[field_name] = message[current_offset:ellipsis_offset]
-                diff_dict[field_name] = True
+                result_dict[field_name_mangled] = message[current_offset:ellipsis_offset]
+                diff_dict[field_name_mangled] = True
                 
                 inserted_length = len(message[current_offset:ellipsis_offset])//2
                                 
@@ -739,10 +832,10 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                 if length_to_insert < 0:
                     number_of_missing_bytes = this_field_expected_length - inserted_length
                     if number_of_missing_bytes > 0:
-                        result_dict[field_name] += f"...({number_of_missing_bytes} bytes)..."
+                        result_dict[field_name_mangled] += f"...({number_of_missing_bytes} bytes)..."
                     
-                    result_without_missing_bytes_info = re.sub("...\(.*\)...", "...", result_dict[field_name])
-                    diff_dict.update({field_name: field_rule_complies_parent([result_without_missing_bytes_info, field_name], rule)})
+                    result_without_missing_bytes_info = re.sub("...\(.*\)...", "...", result_dict[field_name_mangled])
+                    diff_dict.update({field_name_mangled: field_rule_complies_parent([result_without_missing_bytes_info, field_name], rule)})
                     
                     current_offset = ellipsis_offset
                     
@@ -750,27 +843,27 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                     number_of_missing_bytes = this_field_expected_length - (inserted_length + length_to_insert)
                     if number_of_missing_bytes >= 0:
                         if number_of_missing_bytes > 0:
-                            result_dict[field_name] += f"...({number_of_missing_bytes} bytes)..."
-                        result_dict[field_name] += message[after_ellipsis_offset:after_ellipsis_offset+length_to_insert*2]
+                            result_dict[field_name_mangled] += f"...({number_of_missing_bytes} bytes)..."
+                        result_dict[field_name_mangled] += message[after_ellipsis_offset:after_ellipsis_offset+length_to_insert*2]
                         
-                        result_without_missing_bytes_info = re.sub("...\(.*\)...", "...", result_dict[field_name])
-                        diff_dict.update({field_name: field_rule_complies_parent([result_without_missing_bytes_info, field_name], rule)})
+                        result_without_missing_bytes_info = re.sub("...\(.*\)...", "...", result_dict[field_name_mangled])
+                        diff_dict.update({field_name_mangled: field_rule_complies_parent([result_without_missing_bytes_info, field_name_mangled], rule)})
                         
                     else:
                         is_valid = False
-                        diff_dict[field_name] = False
+                        diff_dict[field_name_mangled] = False
                         
                         after_overflow_offset = after_ellipsis_offset+length_to_insert*2 - (-number_of_missing_bytes*2)
                         middle_overflowing_msg = message[after_ellipsis_offset:after_ellipsis_offset+(-number_of_missing_bytes*2)]
                         correct_msg_part = message[after_ellipsis_offset+(-number_of_missing_bytes*2):]
                         
-                        result_dict[field_name] += f"(+{middle_overflowing_msg}) {correct_msg_part}"
-                        log_message = f"Message with ellipsis too long while checking field '{field_name}', overflowing {-number_of_missing_bytes} bytes: '{message[after_ellipsis_offset:after_ellipsis_offset+(-number_of_missing_bytes*2)]}'  after the ellipsis"
+                        result_dict[field_name_mangled] += f"(+{middle_overflowing_msg}) {correct_msg_part}"
+                        log_message = f"Message with ellipsis too long while checking field '{field_name_mangled}', overflowing {-number_of_missing_bytes} bytes: '{message[after_ellipsis_offset:after_ellipsis_offset+(-number_of_missing_bytes*2)]}'  after the ellipsis"
                                                                                                                                                                             
                         try:
-                            log_dict[field_name].append(log_message)
+                            log_dict[field_name_mangled].append(log_message)
                         except KeyError:
-                            log_dict[field_name] = [log_message]
+                            log_dict[field_name_mangled] = [log_message]
                 
             
                     # We should now have sth like this in the result_dict[field_name]:
@@ -784,29 +877,38 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
                 
                 if current_length > 0:
                     aux_rule = [message[current_offset:current_offset+current_length * 2], field_name]
-                    diff_dict.update({field_name: field_rule_complies_parent(aux_rule, rule)})
+                    diff_dict.update({field_name_mangled: field_rule_complies_parent(aux_rule, rule)})
                     
-                    result_dict.update({field_name: message[current_offset:current_offset+current_length * 2]})
+                    result_dict.update({field_name_mangled: message[current_offset:current_offset+current_length * 2]})
                     
                     if field_rule_is_encoded(rule):
-                        decoded_result_dict[field_name] = f"{field_decode(rule, message_field_subtring)}"
+                        decoded_result_dict[field_name_mangled] = f"{field_decode(rule, message_field_subtring)}"
                 else:
                     current_length = 0
                     
-                    diff_dict.update({field_name: False})
-                    result_dict.update({field_name: "--"})
+                    diff_dict.update({field_name_mangled: False})
+                    result_dict.update({field_name_mangled: "--"})
                     
                     log_message = f"expected invalid length of {current_length} for field in rule {rule}"
                     #print(log_message)
                                                                                                                                                                             
                     try:
-                        log_dict[field_name].append(log_message)
+                        log_dict[field_name_mangled].append(log_message)
                     except KeyError:
-                        log_dict[field_name] = [log_message]
+                        log_dict[field_name_mangled] = [log_message]
             
             
                 current_offset += current_length * 2
-        
+            
+        elif rule_is_multifieldstart(rule):
+            multifield_name = get_multifieldstart_full_name(rule[1])
+            multifield_names_stack.append(multifield_name)
+        elif rule_is_multifieldend(rule):
+            mfs_name = multifield_names_stack[-1]
+            mfe_name = get_multifieldend_full_name(rule[1])
+            assert mfs_name == mfe_name, f"Unexpected multifield end, {rule=}, {multifield_names_stack[-1]=}"
+            multifield_names_stack.pop()
+            
         i=i+1
     
     if current_offset < len(message):
@@ -821,7 +923,7 @@ def validate_message(message_rules, message, all_messages_rules_tokenized):
         result_dict.update({None: message[current_offset:]})
         
     #pprint(diff_dict)
-    #pprint(result_dict)
+    # pprint(result_dict)
     #print(is_valid)
     
     if is_valid == True:
@@ -1183,7 +1285,8 @@ def cli_main():
     ret = 0
         
     all_messages_rules_tokenized = [tokenize_rules(r) for r in split_multimessage_rules(all_messages_string)]
-    
+    preprocess_encode_fields(all_messages_rules_tokenized)
+    all_messages_rules_tokenized = preprocess_multiple_subtypeof(all_messages_rules_tokenized)
     
     if args["check"] == True:
         messages_input = [sys.stdin.read()] if not args["<message_hex_string>"] else args["<message_hex_string>"]
@@ -1224,19 +1327,23 @@ def cli_main():
                     if validate_result[0] == False:
                         ret = 1
                 
+                name_string = match
+                if args["--long-name"] == False:
+                    name_string = re.sub(r'.*?([^\.]*)$', r'\1', match)
+                    
                 if args["--list"] == False:
                     if oneline_enabled:
-                        print(color  + f"[{match}]" + AnsiColors.ENDC + "\t" + explanation)
+                        print(color  + f"[{name_string}]" + AnsiColors.ENDC + "\t" + explanation)
                     else:
-                        print(color  + f"[{match}]" + AnsiColors.ENDC)
+                        print(color  + f"[{name_string}]" + AnsiColors.ENDC)
                         print(explanation)
                         print()
                     break
                 else:
                     if oneline_enabled:
-                        print(color  + f"- {i}: [{match}]" + AnsiColors.ENDC + "\t" + explanation)
+                        print(color  + f"- {i}: [{name_string}]" + AnsiColors.ENDC + "\t" + explanation)
                     else:
-                        print(color  + f"- {i}: [{match}]" + AnsiColors.ENDC)
+                        print(color  + f"- {i}: [{name_string}]" + AnsiColors.ENDC)
                         print(explanation)
                         print()
     elif args["create"] == True:
@@ -1290,26 +1397,26 @@ def cli_main():
 
 #TODO warnings: 2 equivalent messages in rules
 #TODO error: 2 fields with same name in rules
-#TODO refactor: organizar rules en clases
 #TODO feature: complete tree in multiline check/dissect
 #TODO ?: identificación certera del mensaje en función del message_type???
 #TODO fix: falla cuando un lengthof cae dentro de una ellipsis o más allá del fin del mensaje en mensajes incompletos
 #TODO improvement: cambiar el --dissect-fields por un arg adicional opcional filter-fields que tb funcione con el check
 #TODO feature: #include message, #includepart message
-#TODO feature: variation of message (1byte,2 byte) -> double subtypeof?
 #TODO feature: X16
 #TODO improvement: N field of missing length could be OK sometimes
 #TODO feature: endswith instead of length
 #TODO feature: --input-format=bin, --input-format=hex-string
 #TODO feature: create message
 #TODO feature: regex matcher for ascii rule
-#TODO doc: Foo Protocol
 #TODO tests: Bash diff tests
 #TODO fix: Logger for --verbose fix
-#TODO fix: collision between parent and child names. store multifield stack in length_params
 #TODO feature: --input-format=json
 #TODO feature: output-format==ptable
 #TODO optimization: don't tokenize rules for each validation
+#TODO fix: overriden fields with different params, like encodedas
+#TODO optimization: if a parent rule fails, don't check subtypes. --list'd not be possible
+#TODO refactor: CocoDocument, CocoMessageSpec, CocoRule, CocoParser, CocoAnalyzer, CocoCLI
+#TODO improvement: cocofile checks: no "." in any rule
 
             
 if __name__ == "__main__":
